#!/usr/local/bin/perl
# ESA (C) 2000-2018 
#
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
#
# dscheck: Perform table-driven checks of a dataset
#
# Version 0.1  rgw  02-Jul-1999  Initial version
#

require 5.005;

# Be strict about everything
use strict;
use vars qw($taskname $VERSION);

# We use the SAS/DAL stuff
use DAL;
use SAS;

# Other standard perl modules and pragmas
use English;
use Text::ParseWords;

sub dscheck {
    $taskname='dscheck';
    $VERSION='0.1';

    # Fetch parameters from command-line
    my $infile=stringParameter("set");
    my $checkfile=stringParameter("checkfile");

    # Open the dataset
    my $ds=DAL::dataset->new(name => $infile, mode => $DAL::READ);

    # Open and read the check file
    my @checks;
    open(FILE,"<$checkfile") or die "Cannot open check file ($OS_ERROR)";
    chomp(@checks=<FILE>);
    close(FILE);

    # Loop through each check, parsing and performing the check
    my ($fail,@out,@log);
    foreach my $check (grep(!/^\#/,@checks)) {
	$check=~s/^\s+//;
	my @words=split(/[, ]+/,$check);
	my $cmd=lc shift @words;

	# Remain calm about up-coming symbolic reference
	no strict 'refs';

	# Check that the function exists, if so call it
	my $func="check_$cmd";
	if(exists $main::{$func}){
	    my ($thisfail,@out)=&{"main::$func"}($ds,@words);
	    push(@log,@out);
	    $fail=1 if $thisfail;
	}
    }
    
    print join("\n",@log,"");

    SAS::fatal('checkFail','One or more checks failed') if $fail;
}

# Parse the specification for a block plus attribute/column
#
sub parse_item {
    my $what=shift;

    my ($block,$attr)=split(/\./,$what);
    my $isprimary=($block=~/primary/i);
    my $iscolumn=($attr=~/\[\]$/);
    SAS::error('primCol','Cannot refer to columns in the primary header')
      if $iscolumn and $isprimary;
    $attr=~s/\[\]//g if $iscolumn;
    return ($block,$attr,$isprimary,$iscolumn);
}


# Check for existence of block, block attribute or table column
#
sub check_exists {
    my $ds=shift;
    my @list=@ARG;

    # Loop over each specified block or block attribute
    my ($fail,@out);
    $fail=0;
    foreach (@list) {
	my $ok=1;		# Assume things are good to start with
	my ($block,$attr,$isprimary,$iscolumn)=parse_item($ARG);

	# Check for the existence of the block
	unless($isprimary) {
	    $ok=$ds->hasBlock(name => $block);
	    push(@out,"Missing block '$block'") unless $ok;
	}

	if($ok) {
	    # Get handle to block
	    my $h=($isprimary)?$ds:$ds->table(name => $block);
	    
	    # Check the attribute/column if needed (and possible)
	    if(defined $attr && $ok) {
		
		# Is this a check on an attribute or column?
		if($iscolumn){	# column check
		    $ok=$h->hasColumn(name => $attr);
		    push(@out,"Missing table column '$block.$attr'") unless $ok;
		}else{
		    $ok=$h->hasAttribute(name => $attr);
		    push(@out,"Missing attribute '$block.$attr'") unless $ok;
		}
	    }
	}

	# Remember failures
	$fail=1 unless $ok;
    }

    # Return
    return ($fail,@out);
}

# Range check on attribute or table column
#
sub check_range {
    my ($ds,$what,$min,$max)=@ARG;

    # Work out what we're checking
    my ($block,$attr,$isprimary,$iscolumn)=parse_item($what);

    # First do an existence check
    my ($fail,@out)=check_exists($ds,$what);

    # If OK, carry on with range checks
    unless($fail){
	my $h=($isprimary)?$ds:$ds->table(name => $block);

	# Read value from keyword or table column, find any bad values
	# This is NOT good-looking perl, due to short-comings of PEDAL
	my ($value,$nbad);
	if($iscolumn){
	    my $co=$h->column(name => $attr);
	    my $vref=$co->data();
	    for(my $i=0;$i<$h->rows;$i++){
		my $value=$vref->value($i);
		$nbad++ if $value<$min or $value>$max;
	    }
	    if($nbad){
		push(@out,"Table column '$block.$attr' contains $nbad out-of-range values");
		$fail=1;
	    }
	}else{
	    my $ath=$value=$h->attribute(name => $attr);
	    $value=$ath->asReal64;
	    if($value<$min or $value>$max) {
		push(@out,"Value of attribute '$what' out-of-range");
		$fail=1;
	    }
	}
    }

    return ($fail,@out);
}

1;
